import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="alfi",
    version="1.0.1",
    author="Jacob Moss",
    author_email="cob.mossy@gmail.com",
    description="An approximate latent force model library with variational inference for non-linear ODEs and PDEs.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mossjacob/lafomo",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        # 'tensorflow>=2.4.1',
        # 'tensorflow-probability>=0.12.1',
        'torch>=1.7.1',
        'torchdiffeq>=0.2.0',
        'pandas>=1.2.1',
        'matplotlib',
        'gpytorch>=1.3.1'
    ]
)



import csv
import pandas as pd
import numpy as np
from sklearn import preprocessing
from os import path


class DataHolder(object):
    def __init__(self, data, noise, time):
        self.m_obs, self.f_obs = data
        if noise is not None:
            self.σ2_m_pre, self.σ2_f_pre = noise
        self.t_observed = time[0]
        self.t_discretised = time[1]
        self.common_indices = time[2]


def load_covid():
    with open('data/covidrld.csv', 'r', 1) as f:
        contents = f.buffer
        df = pd.read_table(contents, sep=',', index_col=0)
    replicates = 2
    columns = df.columns
    #ENSMPUG is the prefix for Ferret
    known_target_genes = [
        # 'ENSMPUG00000000676', #EDN1 
        'ENSMPUG00000002543', #OAS1
        'ENSMPUG00000013592', #IL6
        # 'ENSMPUG00000008689', #CXCL10
        'ENSMPUG00000008778', #Mx-1
        'ENSMPUG00000013925', #ISG15
        'ENSMPUG00000015816', #ISG20
        # 'ENSMPUG00000004319', #RIG-I
    ]
    known_tfs = [
        # 'ENSMPUG00000010522', #IRF1
        # 'ENSMPUG00000008845', #IRF7
        'ENSMPUG00000012530', #STAT1
        'ENSMPUG00000001431', #STAT2
    ]

    genes_df = df[df.index.isin(known_target_genes)][columns]
    tfs_df = df[df.index.isin(known_tfs)][columns]

    #Normalise across time points
    normalised = preprocessing.normalize(np.r_[genes_df.values,tfs_df.values])
    genes = normalised[:genes_df.shape[0]]
    tfs = normalised[genes_df.shape[0]:]

    genes = np.stack([genes[:, [0,2,4,6]], genes[:, [1,3,5,7]]])
    tfs = np.stack([tfs[:, [0,2,4,6]], tfs[:, [1,3,5,7]]])

    return (genes_df, np.float64(genes)), (tfs_df, np.float64(tfs)), np.array([0, 3, 7, 14])

def load_3day_dros():
    with open('data/3day/GSE47999_Normalized_Counts.txt', 'r', 1) as f:
        contents = f.buffer
        df = pd.read_table(contents, sep='\t', index_col=0)
    replicates = 3
    columns = df.columns[df.columns.str.startswith('20,000')][::replicates]
    known_target_genes = ['FBgn0011774', 'FBgn0030189', 'FBgn0031713', 'FBgn0032393', 'FBgn0037020', 'FBgn0051864']
    tf_names = ['FBgn0039044']
    genes_df = df[df.index.isin(known_target_genes)][columns]
    tfs_df = df[df.index.isin(tf_names)][columns]

    #Normalise across time points
    normalised = preprocessing.normalize(np.r_[genes_df.values,tfs_df.values])
    genes = normalised[:-1]
    tfs = np.atleast_2d(normalised[-1])
    return (genes_df, np.float64(genes)), (tfs_df, np.float64(tfs)), np.array([2, 10, 20])


def load_barenco_puma(dir_path):
    mmgmos_processed = True
    if mmgmos_processed:
        with open(path.join(dir_path, 'barencoPUMA_exprs.csv'), 'r') as f:
            df = pd.read_csv(f, index_col=0)
        with open(path.join(dir_path, 'barencoPUMA_se.csv'), 'r') as f:
            dfe = pd.read_csv(f, index_col=0)
        columns = [f'cARP{r}-{t}hrs.CEL' for r in range(1, 4) for t in np.arange(7)*2]
    else:
        with open(path.join(dir_path, 'barenco_processed.tsv'), 'r') as f:
            df = pd.read_csv(f, delimiter='\t', index_col=0)

        columns = [f'H_ARP1-{t}h.3' for t in np.arange(7)*2]

    known_target_genes = ['203409_at', '202284_s_at', '218346_s_at', '205780_at', '209295_at', '211300_s_at']
    genes = df[df.index.isin(known_target_genes)][columns]
    genes_se = dfe[dfe.index.isin(known_target_genes)][columns]

    assert df[df.duplicated()].size == 0

    index = {
        '203409_at': 'DDB2',
        '202284_s_at': 'p21',
        '218346_s_at': 'SESN1',
        '205780_at': 'BIK',
        '209295_at': 'DR5',
        '211300_s_at': 'p53'
    }
    genes.rename(index=index, inplace=True)
    genes_se.rename(index=index, inplace=True)

    # Reorder genes
    genes_df = genes.reindex(['DDB2', 'BIK', 'DR5', 'p21', 'SESN1', 'p53'])
    genes_se = genes_se.reindex(['DDB2', 'BIK', 'DR5', 'p21', 'SESN1', 'p53'])

    tfs_df = genes_df.iloc[-1:]
    genes_df = genes_df.iloc[:-1]
    genes = genes_df.values
    tfs = tfs_df.values

    tf_var = genes_se.iloc[-1:].values**2
    gene_var = genes_se.iloc[:-1].values**2

    tfs_full = np.exp(tfs + tf_var/2)
    genes_full = np.exp(genes+gene_var/2)

    tf_var_full = (np.exp(tf_var)-1)*np.exp(2*tfs + tf_var)
    gene_var_full = (np.exp(gene_var)-1)*np.exp(2*genes + gene_var) # This mistake is in Lawrence et al.

    tf_scale = np.sqrt(np.var(tfs_full[:, :7], ddof=1))
    tf_scale = np.c_[[tf_scale for _ in range(7*3)]].T
    tfs = np.float64(tfs_full / tf_scale).reshape((3, 1, 7))
    tf_var = (tf_var_full / tf_scale**2).reshape((3, 1, 7))

    gene_scale = np.sqrt(np.var(genes_full[:,:7], axis=1, ddof=1))
    gene_scale = np.c_[[gene_scale for _ in range(7*3)]].T
    genes = np.float64(genes_full / gene_scale).reshape((5, 3, 7)).swapaxes(0, 1)
    gene_var = np.float64(gene_var_full / gene_scale**2).reshape((5, 3, 7)).swapaxes(0, 1)

    return (genes_df, genes), (tfs_df, np.float64(tfs)), gene_var, tf_var, np.arange(7)*2           # Observation times


def scaled_barenco_data(f):
    scale_pred = np.sqrt(np.var(f))
    barencof = np.array([[0.0, 200.52011, 355.5216125, 205.7574913, 135.0911372, 145.1080997, 130.7046969],
                         [0.0, 184.0994134, 308.47592, 232.1775328, 153.6595161, 85.7272235, 168.0910562],
                         [0.0, 230.2262511, 337.5994811, 276.941654, 164.5044287, 127.8653452, 173.6112139]])

    barencof = barencof[0]/(np.sqrt(np.var(barencof[0])))*scale_pred
    # measured_p53 = df[df.index.isin(['211300_s_at', '201746_at'])]
    # measured_p53 = measured_p53.mean(0)
    # measured_p53 = measured_p53*scale_pred
    measured_p53 = 0
    
    return barencof, measured_p53


# From barenco paper
def barenco_params():
    return np.stack([
        np.array([2.6, 1.5, 0.5, 0.2, 1.35])[[0, 4, 2, 3, 1]],
        (np.array([1.2, 1.6, 1.75, 3.2, 2.3])*0.8/3.2)[[0, 4, 2, 3, 1]],
        (np.array([3, 0.8, 0.7, 1.8, 0.7])/1.8)[[0, 4, 2, 3, 1]]
    ]).T




import numpy as np

from alfi.trainers import Trainer


class ExactTrainer(Trainer):
    def __init__(self, *args, loss_fn, **kwargs):
        super().__init__(*args, **kwargs)
        self.loss_fn = loss_fn
        self.losses = np.empty((0, 1))

    def single_epoch(self, **kwargs):
        epoch_loss = 0

        [optim.zero_grad() for optim in self.optimizers]
        # Output from model
        output = self.model(self.model.train_t)
        # print(output.mean.shape)
        # plt.imshow(output.covariance_matrix.detach())
        # plt.colorbar()
        # Calc loss and backprop gradients
        loss = -self.loss_fn(output, self.model.train_y.squeeze())
        loss.backward()
        [optim.step() for optim in self.optimizers]
        epoch_loss += loss.item()

        return epoch_loss, [epoch_loss]

    def print_extra(self):
        self.model.covar_module.lengthscale.item(),
        self.model.likelihood.noise.item()
        super().print_extra()



import gpytorch
import torch
import numpy as np

from gpytorch.constraints import Positive, Interval

PI = torch.tensor(np.pi, requires_grad=False)


class SIMKernel(gpytorch.kernels.Kernel):
    """
    Simple Input Module Motif Kernel
    This kernel is the multi-output cross-kernel for linear response to single transcription factor.
    In other words, it constructs a JTxJT matrix where J is num genes and T is num timepoints.
    SIM: Single Input Module motif
    """

    is_stationary = True

    def __init__(self, num_genes, variance, dtype=torch.float32, **kwargs):
        super().__init__(**kwargs)
        self.num_genes = num_genes
        self.pos_constraint = Positive()
        self.lengthscale_constraint = Interval(0.5, 3.5)

        self.register_parameter(
            name='raw_lengthscale', parameter=torch.nn.Parameter(
                self.lengthscale_constraint.inverse_transform(1 * torch.ones(1, 1, dtype=dtype)))
        )
        self.register_parameter(
            name='raw_decay', parameter=torch.nn.Parameter(
                self.pos_constraint.inverse_transform(1 * torch.ones(self.num_genes, dtype=dtype)))
        )
        self.register_parameter(
            name='raw_sensitivity', parameter=torch.nn.Parameter(
                self.pos_constraint.inverse_transform(1 * torch.ones(self.num_genes, dtype=dtype)))
        )
        self.register_parameter(
            name='raw_noise', parameter=torch.nn.Parameter(1 * torch.ones(self.num_genes, dtype=dtype))
        )

        # register the constraints
        self.register_constraint("raw_lengthscale", self.lengthscale_constraint)
        self.register_constraint("raw_sensitivity", Positive())
        self.register_constraint("raw_decay", Positive())
        self.register_constraint("raw_noise", self.pos_constraint)

        self.variance = torch.diag(variance)

    @property
    def lengthscale(self):
        return self.lengthscale_constraint.transform(self.raw_lengthscale)

    @lengthscale.setter
    def lengthscale(self, value):
        self.initialize(raw_lengthscale=self.lengthscale_constraint.inverse_transform(value))

    @property
    def decay(self):
        return self.pos_constraint.transform(self.raw_decay)

    @decay.setter
    def decay(self, value):
        self.initialize(raw_decay=self.pos_constraint.inverse_transform(value))

    @property
    def sensitivity(self):
        return self.pos_constraint.transform(self.raw_sensitivity)

    @sensitivity.setter
    def sensitivity(self, value):
        self.initialize(raw_sensitivity=self.pos_constraint.inverse_transform(value))

    @property
    def noise(self):
        return self.pos_constraint.transform(self.raw_noise)

    @noise.setter
    def noise(self, value):
        self.initialize(raw_noise=self.pos_constraint.inverse_transform(value))

    def plot_cov(self, x1, x2):
        from matplotlib import pyplot as plt
        Kxx = self(x1, x2)
        plt.figure()
        plt.imshow(Kxx.detach().evaluate().detach())
        plt.colorbar()
        return Kxx

    def forward(self, t1, t2, **params):
        """
        This calculates Kxx (not cross-covariance)
        t1, t2 are blocked timepoint vectors
        Parameters:
           x1 shape (num_genes*num_times)
        """
        vert_block_size = int(t1.shape[0] / self.num_genes)
        hori_block_size = int(t2.shape[0] / self.num_genes)
        t1_block, t2_block = t1[:vert_block_size], t2[:hori_block_size]
        shape = [vert_block_size * self.num_genes, hori_block_size * self.num_genes]
        K_xx = torch.zeros(shape, dtype=torch.float64)
        for j in range(self.num_genes):
            for k in range(self.num_genes):
                kxx = self.k_xx(j, k, t1_block, t2_block)
                K_xx[j * vert_block_size:(j + 1) * vert_block_size,
                     k * hori_block_size:(k + 1) * hori_block_size] = kxx

        if hori_block_size == vert_block_size:
            # noise = self.noise.view(-1, 1).repeat(1, hori_block_size).view(-1)
            # noise = torch.diag(noise)
            jitter = 1e-4 * torch.eye(K_xx.shape[0], dtype=torch.float64)
            K_xx += jitter
            if K_xx.shape[0] == self.variance.shape[0]:
                K_xx += self.variance
        return K_xx

    def k_xx(self, j, k, t1_block, t2_block):
        """
        k_xx(t, t')
        t2 = t'
        Parameters:
            t1_block: tensor shape (T1,)
            t2_block: tensor shape (T2,)
        """
        t1_block = t1_block.view(-1, 1)
        t2_block = t2_block.view(1, -1)
        mult = self.sensitivity[j] * self.sensitivity[k] * self.lengthscale * 0.5 * torch.sqrt(PI)
        k_xx = mult * (self.h(k, j, t2_block, t1_block) + self.h(j, k, t1_block, t2_block))
        return k_xx

    def h(self, k, j, t2, t1):
        l = self.lengthscale
        t_dist = t2 - t1
        multiplier = torch.exp(self.gamma(k) ** 2) / (self.decay[j] + self.decay[k])  # (1, 1)
        first_erf_term = torch.erf(t_dist / l - self.gamma(k)) + torch.erf(t1 / l + self.gamma(k))  # (T,T)
        second_erf_term = torch.erf(t2 / l - self.gamma(k)) + torch.erf(self.gamma(k))
        return multiplier * (torch.multiply(torch.exp(-self.decay[k] * t_dist), first_erf_term) - \
                             torch.multiply(torch.exp(-self.decay[k] * t2 - self.decay[j] * t1), second_erf_term))

    def gamma(self, k):
        return self.decay[k] * self.lengthscale / 2

    def K_xf(self, x, f):
        """
        K_xf
        Cross-covariance. Not optimised (not in marginal likelihood).
        Parameters:
            x: tensor (JT, JT) the blocked observation vector
            f: tensor (T*) the non-blocked f prediction timepoint vector
        """
        shape = [x.shape[0], f.shape[0]]
        K_xf = torch.zeros(shape, dtype=torch.float32)
        self.block_size = int(x.shape[0] / self.num_genes)  # 7
        t1_block, t2_block = x[:self.block_size].view(-1, 1), f.view(1, -1)
        for j in range(self.num_genes):
            kxf = self.k_xf(j, t1_block, t2_block)
            K_xf[j * self.block_size:(j + 1) * self.block_size] = kxf

        return K_xf

    def k_xf(self, j, x, t_f):
        l = self.lengthscale
        t_dist = x - t_f
        erf_term = torch.erf(t_dist / l - self.gamma(j)) + torch.erf(t_f / l + self.gamma(j))
        return self.sensitivity[j] * l * 0.5 * torch.sqrt(PI) * torch.exp(self.gamma(j) ** 2) * torch.exp(
            -self.decay[j] * t_dist) * erf_term

    def K_ff(self, x1, x2):
        """Returns the RBF kernel between latent TF"""
        x1 = x1.view(-1)
        x2 = x2.view(-1)
        sq_dist = torch.square(x1.view(-1, 1)-x2)
        sq_dist = torch.div(sq_dist, 2*self.lengthscale.view((-1, 1)))
        K = torch.exp(-sq_dist)
        if K.shape[0] == K.shape[1]:
            jitter = 1e-3 * torch.eye(x1.shape[0])
            K += jitter
        return K



import torch
import gpytorch

from torch.distributions import MultivariateNormal

from .lfm import LFM
from alfi.kernels import SIMKernel
from alfi.means import SIMMean
from alfi.utilities.data import flatten_dataset
from alfi.datasets import LFMDataset


class ExactLFM(LFM, gpytorch.models.ExactGP):
    """
    An implementation of the single input motif from Lawrence et al., 2006.
    """
    def __init__(self, dataset: LFMDataset, variance):
        train_t, train_y = flatten_dataset(dataset)
        super().__init__(train_t, train_y, likelihood=gpytorch.likelihoods.GaussianLikelihood())

        self.num_outputs = dataset.num_outputs
        self.train_t = train_t.view(-1, 1)
        self.train_y = train_y.view(-1, 1)
        self.covar_module = SIMKernel(self.num_outputs, torch.tensor(variance, requires_grad=False))
        initial_basal = torch.mean(train_y.view(self.num_outputs, -1), dim=1) * self.covar_module.decay
        self.mean_module = SIMMean(self.covar_module, self.num_outputs, initial_basal)

    @property
    def decay_rate(self):
        return self.covar_module.decay

    @decay_rate.setter
    def decay_rate(self, val):
        self.covar_module.decay = val

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)

    def predict_m(self, pred_t, jitter=1e-5) -> torch.distributions.MultivariateNormal:
        """
        Predict outputs of the LFM
        :param pred_t: prediction times
        :param jitter:
        :return:
        """
        Kxx = self.covar_module(self.train_t, self.train_t)
        K_inv = torch.inverse(Kxx.evaluate())
        pred_t_blocked = pred_t.repeat(self.num_outputs)
        K_xxstar = self.covar_module(self.train_t, pred_t_blocked).evaluate()
        K_xstarx = torch.transpose(K_xxstar, 0, 1).type(torch.float64)
        K_xstarxK_inv = torch.matmul(K_xstarx, K_inv)
        KxstarxKinvY = torch.matmul(K_xstarxK_inv, self.train_y)
        mean = KxstarxKinvY.view(self.num_outputs, pred_t.shape[0])

        K_xstarxstar = self.covar_module(pred_t_blocked, pred_t_blocked).evaluate()
        var = K_xstarxstar - torch.matmul(K_xstarxK_inv, torch.transpose(K_xstarx, 0, 1))
        var = torch.diagonal(var, dim1=0, dim2=1).view(self.num_outputs, pred_t.shape[0])
        mean = mean.transpose(0, 1)
        var = var.transpose(0, 1)
        var = torch.diag_embed(var)
        var += jitter * torch.eye(var.shape[-1])
        return MultivariateNormal(mean, var)

    def predict_f(self, pred_t, jitter=1e-3) -> MultivariateNormal:
        """
        Predict the latent function.

        :param pred_t: Prediction times
        :param jitter:
        :return:
        """
        Kxx = self.covar_module(self.train_t, self.train_t)
        K_inv = torch.inverse(Kxx.evaluate())

        Kxf = self.covar_module.K_xf(self.train_t, pred_t).type(torch.float64)
        KfxKxx = torch.matmul(torch.transpose(Kxf, 0, 1), K_inv)
        mean = torch.matmul(KfxKxx, self.train_y).view(-1).unsqueeze(0)

        #Kff-KfxKxxKxf
        Kff = self.covar_module.K_ff(pred_t, pred_t)  # (100, 500)
        var = Kff - torch.matmul(KfxKxx, Kxf)
        # var = torch.diagonal(var, dim1=0, dim2=1).view(-1)
        var = var.unsqueeze(0)
        # For some reason a full covariance is not PSD, for now just take the variance: (TODO)
        var = torch.diagonal(var, dim1=1, dim2=2)
        var = torch.diag_embed(var)
        var += jitter * torch.eye(var.shape[-1])

        batch_mvn = gpytorch.distributions.MultivariateNormal(mean, var)
        return gpytorch.distributions.MultitaskMultivariateNormal.from_batch_mvn(batch_mvn, task_dim=0)

    def save(self, filepath):
        torch.save(self.state_dict(), filepath+'lfm.pt')

    @classmethod
    def load(cls,
             filepath,
             lfm_args=[], lfm_kwargs={}):
        lfm_state_dict = torch.load(filepath+'lfm.pt')
        lfm = cls(*lfm_args, **lfm_kwargs)
        lfm.load_state_dict(lfm_state_dict)
        return lfm



from .base import Trainer
from .exact import ExactTrainer
from .variational import VariationalTrainer
from .preestimator import PreEstimator, PartialPreEstimator
from .operator import NeuralOperatorTrainer
try:  # fenics may not be present
    from .partial import PDETrainer
except ImportError:
    pass


__all__ = [
    'Trainer',
    'ExactTrainer',
    'VariationalTrainer',
    'PDETrainer',
    'PreEstimator',
    'PartialPreEstimator',
    'NeuralOperatorTrainer',
]



import torch
from torch import nn
import torch.nn.functional as F
import pytorch_lightning as pl


class MLP(pl.LightningModule):
    def __init__(self, input_dim, output_dim, latent_dim=None, num_hidden_layers=3):
        super().__init__()
        latent_dim = 2 * ((input_dim + output_dim) // 2) if latent_dim is None else latent_dim
        middle_layers = list()
        for i in range(num_hidden_layers):
            middle_layers.extend([
                nn.Linear(latent_dim, latent_dim),
                nn.ReLU(),
            ])
        self.network = nn.Sequential(
            nn.Linear(input_dim, latent_dim),
            nn.ReLU(),
            *middle_layers,
        )
        self.output_head = nn.Sequential(
            nn.Linear(latent_dim, output_dim)
        )
        self.latent_head = nn.Sequential(
            nn.Linear(latent_dim, output_dim)
        )

    def forward(self, x, use_output_head=True):
        x = self.network(x)
        # if use_output_head:
        return self.output_head(x)

    def training_step(self, batch, batch_idx):
        x, y = batch
        output_tensor = self(x)
        mse = F.mse_loss(output_tensor.squeeze(), y)
        self.log("train_loss", mse.item())
        return mse

    def validation_step(self, batch, batch_idx):
        x, y = batch
        output_tensor = self(x)
        mse = F.mse_loss(output_tensor.squeeze(), y)
        self.log("val_loss", mse.item())
        return mse

    def configure_optimizers(self):
        return torch.optim.Adam(self.parameters())
